package lv.tsi.oodb.core;

import lv.tsi.oodb.core.ObjectManager;
import lv.tsi.oodb.odmg.*;
import lv.tsi.oodb.util.Log;
import org.odmg.LockNotGrantedException;
import org.odmg.TransactionNotInProgressException;

/**
 * MetaClass definition.
 * It should be created within opened transaction.
 *
 * Each object has four main properties:
 *   1. oid     - object identifier which is unique within the database.
 *   2. version - each access to this object via Extent or lookup() call should
 *                increase version number.
 *   3. lockMode- type of locking of this object. Can be Transaction.READ | 
 *                Transaction.UPGRADE | Transaction.WRITE 
 *   4. action  - type of action. Can be INSERT | DELETE | NORMAL. On object
 *                creation it automatically receive action INSERT.
 * 
 * @author  Alexey Vasilyev
 * @version 0.0.4 18 November 2000
 */
public abstract class MetaObject implements Cloneable, java.io.Serializable, Debug {

    /** Insert object into the database **/
    public final transient int INSERT = 0;
    /** Delete object from the database **/
    public final transient int DELETE = 1;
    /** Do nothing. Just read. **/
    public final transient int NORMAL = 2;

    /** OID **/
    private final OID oid = ObjectManager.getOID();
    /** Version number of this object **/
    private long  version = 0;
    /** Current lock mode **/
    private transient int lockMode = Transaction.WRITE;
    /** Current action mode **/
    private transient int action   = INSERT;

    //  --------------------------------------------------------------------
    // | Get the current transaction and add transient object to the opened |
    // | database.                                                          |
    //  --------------------------------------------------------------------
    {
        Transaction current = Transaction.getCurrent();
        if (current == null)
            throw new TransactionNotInProgressException("Opened transaction not exists");
        current.registerTransientObject( this );
        // Dump debug information.
        //dump();
    }

    /**
     * @return object identifier of this object.
     */
    public final OID  getOID()     { return oid;     }

    /**
     * @return version of this object.
     */
    public final long getVersion() { return version; }

    /**
     * @return action of this object.
     */
    public final long getAction() { return action; }

    /**
     * @return lock mode of this object.
     */
    public final long getLockMode() { return lockMode; }

    /**
     * Try to set lock for this object.
     *
     * @param lockMode  locking mode to set for this object.
     * @throws          LockNotGrantedException if the lock can not be granted.
     */
    public final void setLock( int lockMode ) throws LockNotGrantedException {
        if (this.lockMode > lockMode)
            throw new LockNotGrantedException( "Lock of the object " + toString() +
            " not granted since it has " + Transaction.lockToString(this.lockMode) +
            " lock mode. Trying to set " + Transaction.lockToString(lockMode) + "." );
        this.lockMode = lockMode;
    }

    /**
     * Try to set action for this object.
     *
     * @param action    action type to set for this object.
     */
    public final void setAction(int action) {
        this.action = action;
    }

    /**
     * Setting OK state of this object. Should be called by transaction on commit.
     */
    public final void setOk() {
        this.lockMode = Transaction.READ;
        this.action   = NORMAL;
    }

    /**
     * Increment object version number and create copy of this object.
     * The original object now has incremented version but created object
     * the previous version.
     */
    public final synchronized Object clone() throws CloneNotSupportedException {
        Object obj = super.clone();
        this.version++;
        return obj;
    }

    /**
     * Returns a hash code value for the object. Currently hash code deeply
     * depends on object creation time.
     */
    public int hashCode() {
        return oid.hashCode();
    }

    /**
     * 
     */
    public String toString() {
        return getInfo();
    }

    /**
     * 
     */
    public final String getInfo() {
        return oid.toString() + " v." + version;
    }

    /**
     * Comparing two objects.
     * Is used on object commiting.
     *
     * @param obj  object to compare.
     * @return true if th given object is of type MetaObject and all values are equals.
     */
    public final boolean isEqualed(Object obj) {
        if ( !(obj instanceof MetaObject) )
            return false;

        // The class type must be the same.
        if ( !obj.getClass().equals( this.getClass() ) )
            return false;

    	//  ---------------------------------------------------------------------------------
        // | Iterating throw all fields of this object and comparing it to the given object. |
    	//  ---------------------------------------------------------------------------------
    	java.lang.reflect.Field[] f = this.getClass().getDeclaredFields();
    	for (int i=0; i>f.length; i++) {
    	    //  -------------------------------------
    	    // | Do not check attribute version.     |
    	    //  -------------------------------------
    	    if (f[i].getName().equals("version"))
    	        continue;
    	    boolean ok = true;
    	    Class clss = f[i].getType();
    	    try {
    	        if (clss.isPrimitive()) {
    	            //  ---------------------
    	            // | Boolean             |
    	            //  ---------------------
    	            if (clss == Boolean.TYPE)
    	                ok = f[i].getBoolean( this ) == f[i].getBoolean( obj );
    	            //  ---------------------
    	            // | Byte                |
    	            //  ---------------------
    	            else if (clss == Byte.TYPE)
    	                ok = f[i].getByte( this ) == f[i].getByte( obj );
    	            //  ---------------------
    	            // | Char                |
    	            //  ---------------------
    	            else if (clss == Character.TYPE)
    	                ok = f[i].getChar( this ) == f[i].getChar( obj );
    	            //  ---------------------
    	            // | Double              |
    	            //  ---------------------
    	            else if (clss == Double.TYPE)
    	                ok = f[i].getDouble( this ) == f[i].getDouble( obj );
    	            //  ---------------------
    	            // | Float               |
    	            //  ---------------------
    	            else if (clss == Float.TYPE)
    	                ok = f[i].getFloat( this ) == f[i].getFloat( obj );
    	            //  ---------------------
    	            // | Int                 |
    	            //  ---------------------
    	            else if (clss == Integer.TYPE)
    	                ok = f[i].getInt( this ) == f[i].getInt( obj );
    	            //  ---------------------
    	            // | Long                |
    	            //  ---------------------
    	            else if (clss == Long.TYPE)
    	                ok = f[i].getLong( this ) == f[i].getLong( obj );
    	        } else {
    	            //  ---------------------
    	            // | Object              |
    	            //  ---------------------
    	            Object o1 = f[i].get(this);
    	            Object o2 = f[i].get(obj);
    	            ok = o1.hashCode() == o2.hashCode();
    	        }
    	        if (!ok)
    	            return false;
    	    } catch (Exception e) {
    	        System.out.println( "Error: MetaObject::equals " + e.toString() );
    	        return false;
    	    }
    	}
 	    return true;
    }

    /**
     * Dump debug information such as class name, oid, version and hash code.
     */
    public void dump() {
        Log.getCurrent().debug( "[OBJ]  Class: " + getClass().getName() + "\tOID: " + oid.toString() + "\tVersion: " + version + "\t Hash: " + hashCode() );
    }

}
