package lv.tsi.oodb.core;

import lv.tsi.oodb.odmg.*;
import lv.tsi.oodb.util.Log;
import org.odmg.*;

import java.util.*;

/**
 * Object manager core component implementation.
 * Handles registry of allocated objects.
 *
 * @author  Alexey Vasilyev
 * @version 0.0.4 19 November 2000
 */
public class ObjectManager implements Debug {

    /** Descriptor hashtable of root persistent objects.
        Key=OID, value=instance of ObjectEntry class **/
    private HashMap root = new HashMap(50);

    /** Descriptor hashtable of temp transient objects.
        Key=ClassName, value=instance of ObjectEntry class **/
    private HashMap temp = new HashMap(20);

    /** Descriptor hashtable of root persistent objects.
        Key=ClassName, value=set of instances of ObjectEntry class **/
    //private HashMap extents = new HashMap(50);

    /** Descriptor hashtable of root persistent named objects.
        Key=ClassName, value=instance of ObjectEntry class **/
    private HashMap root_named = new HashMap(50);

    /** Descriptor hashtable of root persistent named objects.
        Key=ClassName, value=instance of ObjectEntry class **/
    private TransientNamedHashMap temp_named = new TransientNamedHashMap();

    /**
     * Create an empty object manager with all empty tables.
     */
    public ObjectManager() {
    }

    /**
     * Adds transient object to the object manager.
     * Is called automatically on any MetaObject class creation within opened
     * transaction.
     *
     * @param object  object which should be added.
     * @throws        ObjectNotUniqueException if the persistent object
     *                with this OID already exist in this database.
     * @throws        OODBRuntimeException if the object is not of type
     *                MetaObject.
     */
    public synchronized void addTransientObject(MetaObject object) throws ObjectNotUniqueException {
        MetaObject meta = null;
        //  ------------------------------------------------------------------------------------
        // | If object with this OID already exists in root table and it's version is smaller   |
        // | than the inserting object                                                          |
        // | or if object with this OID already exists in temp table and it's version           |
        // | equals to 0 throw error                                                            |
        //  ------------------------------------------------------------------------------------
        if (
             (meta = (MetaObject)root.get( object.getOID() )) != null
              &&
              object.getVersion() > meta.getVersion() 
            ||
             (meta = (MetaObject)temp.get( object.getOID() )) != null
              &&
              object.getVersion() == 0
            )
            throw new ObjectNotUniqueException( "Object OID:'"
                + object.getInfo() + "' already exists." );
        // Putting into the transient objects table
        temp.put( object.getOID(), object );
    }

    /**
     *
     * Is called automatically within opened transaction on abort.
     *
     * @param object  the transient object to remove.
     * @throws        ObjectNotFoundException if the object of the
     *                specified name doesn't exist in this database.
     */
    public synchronized void removeTransientObject(MetaObject object) throws ObjectNotFoundException {
        if (temp.get(object.getOID()) == null)
            throw new ObjectNotFoundException( "Transient object OID:'"
                + object.getInfo() + "' not found." );
        // Remove object from transient objects table
        temp.remove( object.getOID() );
        // Trying to remove transient named binded objects from table
        // In most cases this will return null.
        temp_named.remove( object );
    }

    /**
     * Add persistent object to the root descriptor hashtable.
     * Is called on commit.
     *
     * @param object  transient object to persistently commit to the root table.
     * @throws        ObjectNotUniqueException if the object of the
     *                specified OID already exist in this database.
     * @throws        ObjectNotFoundException if the object of the
     *                specified OID doesn't exist in this database.
     */
    public synchronized void addPersistentObject(MetaObject object) throws ObjectNotUniqueException, ObjectNotFoundException {
        MetaObject orig = null;
        if (
             (orig = (MetaObject)root.get( object.getOID() )) != null
            &&
             orig.getVersion() <= object.getVersion()
            )
            throw new ObjectNotUniqueException( "Commit failed. Object OID:'"
                + object.getInfo() + "' already exists in persistent table." );
        if (temp.get(object.getOID()) == null)
            throw new ObjectNotFoundException( "Commit failed. Transient object OID:'"
                + object.getInfo() + "' not found." );

        // If it exists in root table, so we need to delete if it's changed
        if (orig != null && !orig.isEqualed( object ) ) {
            root.remove( orig.getOID() );
            object.setLock( org.odmg.Transaction.WRITE );
        }

        // Putting into the root table
        root.put( object.getOID(), object );

        // Trying to remove transient named binded objects from table
        // In most cases this will return null.
        NamedWrapper wrap = (NamedWrapper)temp_named.remove( object );
        if (wrap != null) {
            // Object can be cloned. So we need to refresh reference.
            wrap = new NamedWrapper( object, wrap.getName() );
            root_named.put( wrap.toString(), wrap );
            //root_named.put( wrap, wrap );
        }
    }

    /**
     * Add persistent object to the root descriptor hashtable.
     * Is called by transaction manager if commit fails.
     */
    public synchronized void removePersistentObject(MetaObject object) throws ObjectNotFoundException {
        if (root.get(object.getOID()) != null)
            throw new ObjectNotFoundException( "Object OID:'"
                + object.getInfo() + "' doesn't exist in persistent table." );
        // Remove object from persistent objects table
        root.remove( object.getOID() );
    }

    /**
     * Bind the named object for easy retrival.
     * Should be called manually by user within opened transaction.
     *
     * @param object  object which should be binded. Must be of type MetaObject
     *                otherwise OODBRuntimeException will be thrown.
     * @param name    string for name.
     * @throws        ObjectNameNotUniqueException if specified name already
     *                exists.
     * @throws        OODBRuntimeException if the object is not of type
     *                MetaObject.
     */
    public synchronized void bind(MetaObject object, String name) throws ObjectNameNotUniqueException {
        if (name == null)
            return;
        if (root_named.get(name) != null || temp_named.get(name) != null)
            throw new ObjectNameNotUniqueException( "Object '"
                + name + "' already exists in named table." );
        temp_named.put( new NamedWrapper(object,name), name );
    }

    /**
     * Get the persistent named object by it's name.
     * Should be called manually by user within opened transaction.
     * Creates copy of this object with incremented version.
     *
     * @param name    name of the binded object.
     * @return        an instance of the persistent object.
     * @throws        ObjectNameNotFoundException if the named object of the
     *                specified name doesn't exist in this database.
     */
    public synchronized MetaObject lookup(String name) throws ObjectNameNotFoundException {
        NamedWrapper obj = (NamedWrapper)root_named.get(name);
        if (obj == null)
            throw new ObjectNameNotFoundException("Persistent named object '" +
                name + "' doesn't exist." );
        MetaObject meta = null;
        try {
            meta = (MetaObject)obj.getObject().clone();
        } catch (CloneNotSupportedException e) {
        }
        return meta;
    }

    /**
     * Disassociate a name with an object. It doesn't remove a persistent
     * object from database.
     *
     * @param name    name of the binded object.
     * @throws        ObjectNameNotFoundException if the named object of the
     *                specified name doesn't exist in this database.
     */
    public synchronized void unbind(String name) throws ObjectNameNotFoundException {
        if (name == null)
            return;
        if (root_named.get(name) == null || temp_named.get(name) == null)
            throw new ObjectNameNotFoundException( "Object '"
                + name + "' doesn't exist in named table." );
        temp_named.remove( name );
    }

    /**
     */
    public synchronized int getRootSize() {
        return root.size();
    }

    /**
     */
    public synchronized int getNamedSize() {
        return root_named.size();
    }

    /**
     */
    synchronized HashMap getRoot() {
        return root;
    }

    /**
     */
    synchronized HashMap getNamed() {
        return root_named;
    }

    /**
     * Get persistent objects of the given class.
     */
    public synchronized Vector getPersistentObjects(String className) {
        // This method implementation is not effictive
        Vector v = new Vector();
        Iterator i = root.values().iterator();
        while (i.hasNext()) {
            MetaObject o = (MetaObject)i.next();
            if (o.getClass().getName().equals( className ))
                try {
                    v.addElement( o.clone() );
                } catch (CloneNotSupportedException e) {}
        }
        return v;
    }

    /**
     * Print debug information such as number of persistent, named and
     * transient objects.
     */
    public void dump() {
        Log.getCurrent().debug( "[OMGR] Persistent#: " + root.size() +
            "\tTransient#: " + temp.size() + 
            "\tPersistent named#: " + root_named.size() + 
            "\tTransient named#: "  + temp_named.size() );
    }

///////////////////////////
// Static methods
//
    /**
     * Get object id (OID).
     * Is called automatically on any MetaObject class creation within opened
     * transaction.
     */
    protected synchronized static OID getOID() {
        long time = System.currentTimeMillis();
        int  salt = 0;
        if (time == prev_time) {
            prev_salt++;
            salt = prev_salt;
        }
        return new OID(time, salt);
    }

    /** The creation time of the previous object **/
    private static long prev_time;
    /** The creation salt value of the previous object which creation time
        equals to the current object time **/
    private static int  prev_salt;

}
