package lv.tsi.oodb.core;

import lv.tsi.oodb.odmg.*;
import lv.tsi.oodb.OODB;

import java.io.*;
import java.util.Iterator;

/**
 * Class for dispatching disk I/O operations.
 *
 * @author  Alexey Vasilyev
 * @version 0.0.6 3 December 2000
 */
public class ObjectSerializer implements Debug {

    private static final String DB_EXT = ".db";
    private static final String FILE_HEADER = "OODB";

    /**
     * Create database object and create new empty object manager.
     */
    public ObjectSerializer() {
    }

    /**
     *
     * @param manager    object manager from where we must save persistent
     *                   objects.
     * @param name       the name of the database (or it's url).
     * @param accessMode the access mode, which should be one of the static
     *                   fields: OPEN_READ_ONLY, OPEN_READ_WRITE, or
     *                   OPEN_EXCLUSIVE.
     * @throws           DatabaseNotFoundException if the database doesn't
     *                   exist.
     */
    public void save(Database db) throws OODBException {
        try {
            // The previous will be overwritten
            ObjectOutputStream oos = new ObjectOutputStream(
                new FileOutputStream(db.getName() + DB_EXT)
            );
            // Setting header of the file
            oos.write(FILE_HEADER.getBytes());

            // Saving all persistent objects
            Iterator i = db.getObjectManager().getRoot().values().iterator();
            while (i.hasNext()) {
                oos.writeObject(i.next());
            }
            // Saving all named persistent objects
            i = db.getObjectManager().getNamed().values().iterator();
            while (i.hasNext()) {
                oos.writeObject(i.next());
            }

            oos.close();
        } catch (IOException e) {
            throw new OODBException( e.toString() );
        }
    }

    /**
     * Load all object via transaction mechanism.
     *
     * @throws           OODBException if the database doesn't
     *                   exist.
     */
    public void load(Database db) throws OODBException {
        try {
            ObjectInputStream ois = new ObjectInputStream(
                new FileInputStream(db.getName() + DB_EXT)
            );
            // Checking header of the file
            byte[] header = new byte[FILE_HEADER.length()];
            ois.read(header, 0, header.length);
            if (!new String(header).equals( FILE_HEADER ))
                throw new OODBException( "File '" + db.getName() + DB_EXT
                    + "' isn't correct. " + new String(header) + FILE_HEADER);

            // Inserting all serialized objects only within opened transaction
            Transaction txn = new Transaction(db);
            txn.begin();
            while (true) {
                try {
                    // trying to read objects from stream until EOFException occurs
                    Object obj = ois.readObject();
                    if (obj instanceof MetaObject)
                        txn.registerTransientObject( (MetaObject)obj );
                    else if (obj instanceof NamedWrapper) {
                        NamedWrapper wr = (NamedWrapper)obj;
                        db.bind( wr.getObject(), wr.getName() );
                    }
                } catch (EOFException e) {
                    break;
                }
            }
            txn.commit();
            ois.close();

        } catch (FileNotFoundException e) {
            // do nothing
            // new file at the database close will be created
            OODB.getCurrent().getLog().warn( "[DB]   The new database will be created. "
                + e.toString() );
        } catch (Exception e) {
            throw new OODBException( e.toString() );
        }
    }

    /**
     * Debug implementation.
     * Dump number of name binded objects, transient and persistent objects.
     */
    public void dump() {
        //objmanager.dump();
    }

}