package lv.tsi.oodb.core;

import lv.tsi.oodb.odmg.*;
import lv.tsi.oodb.util.Log;
import org.odmg.ObjectNameNotUniqueException;
import org.odmg.TransactionInProgressException;
import org.odmg.TransactionNotInProgressException;

import java.util.*;

/**
 * Object manager core component implementation.
 * Handles registry of allocated objects.
 * Is used by HashMap
 *
 * @author  Alexey Vasilyev
 * @version 0.0.2 5 November 2000
 */
public class TransactionHashMap implements Debug {

    private final static int HASH_INIT = 20;

    /** The set of currently running threads.
        Key=Transaction, Value=HashSet where Value=Thread **/
    private HashMap threads      = new HashMap( HASH_INIT );
    /** Table of currently running transactions. Key=Thread, Value=Transaction **/
    private HashMap transactions = new HashMap( HASH_INIT );

    public TransactionHashMap() {
    }

    /**
     * Register transaction in the transaction table.
     * Is called from (Transaction)txn.begin() and (Transaction)txn.join()
     *
     * @param th  thread where transaction exists
     * @param tr  opened transaction to bind
     */
    public synchronized void put( Thread th, Transaction tr ) throws TransactionInProgressException {
        if ( transactions.get( th ) != null )
            throw new TransactionInProgressException( "Opened transaction still exits. You can't do nested transactions. There can be only one opened transaction per a thread." );
        transactions.put( th, tr );
        HashSet set = (HashSet)threads.get( tr );
        if ( set == null ) {
            set = new HashSet( 3 );
            set.add( th );
            threads.put( tr, set );
        } else
            set.add( th );
    }

    /**
     * Remove thread from the transaction and thread tables.
     * Is called from (Transaction)txn.leave()
     *
     * @param th  thread to remove
     */
    public synchronized void remove( Thread th ) throws TransactionInProgressException {
        Transaction tr = (Transaction)transactions.remove( th );
        if ( tr == null )
            throw new TransactionNotInProgressException( "Transaction doesn't exist." );
        ((HashSet)threads.get( tr )).remove( th );
    }

    /**
     * Remove transaction from the transaction and thread tables.
     * Is called from (Transaction)txn.commit(), (Transaction)txn.abort()
     *
     * @param tr  transaction to remove
     */
    public synchronized void remove( Transaction tr ) throws TransactionInProgressException {
        HashSet set = (HashSet)threads.remove( tr );
        if ( set == null )
            throw new TransactionNotInProgressException( "Thread of the transaction doesn't exist." );
        Iterator i = set.iterator();
        while ( i.hasNext() )
            transactions.remove( (Thread)i.next() );
    }

    /**
     * @return currently opened transaction
     */
    public Transaction getTransaction( Thread th ) {
        return (Transaction)transactions.get( th );
    }

    /**
     * Debug implementation.
     */
    public void dump() {
        Log.getCurrent().debug( "[TXNS]\tThreads#: " + threads.size() + "\tTransactions#: " + transactions.size() );
    }

}
