package lv.tsi.oodb.odmg;

import lv.tsi.oodb.core.*;
import lv.tsi.oodb.odmg.*;
import lv.tsi.oodb.OODB;
import org.odmg.*;

import java.util.HashMap;

/**
 * Database ODMG 3.0 implementation.
 *
 * Databases must be opened before starting any transactions that use the
 * database and closed after ending these transactions. 
 * A database application generally begins processing by accessing one or more
 * critical objects and proceeding from there. These objects are root objects,
 * because they lead to interconnected webs of other objects. The ability to
 * name an object (using method bind) and retrieve it later by that name
 * (using method lookup facilitates this start-up capability. A name is not
 * explicitly defined as an attribute of an object. Naming an object also makes
 * it persistent.
 *
 * There are two types of object retrival:
 *   1. By it's binded name.
 *   2. By using extents.
 * There isn't any possibilities to retrive object by using object id (OID).
 *
 * There is a single flat name scope per database; thus all names and OID in a
 * particular database are unique.
 *
 * Each database consists of two types of objects:
 *   1. Persistent - root objects which already commited.
 *   2. Transient  - temporary objects which is not commited. Opened
 *                   transactions still use them.
 *
 * @author  Alexey Vasilyev
 * @version 0.0.4 18 November 2000
 */
public class Database implements org.odmg.Database, Debug {

    /** Access mode. Can be NOT_OPEN | OPEN_READ_ONLY | OPEN_READ_WRITE | OPEN_EXCLUSIVE **/
    private int accessMode;

    /** The name or url of the database **/
    private String name;

    /** Hashtable of the name binded objects **/
    //private HashMap namebinded;

    /** Object Manager **/
    private ObjectManager objmanager;

    /**
     * Create database object and create new empty object manager.
     */
    public Database() {
        accessMode = NOT_OPEN;
        //namebinded = new HashMap(25);
        objmanager = new ObjectManager();
    }

    /**
     * Check does the current database is opened.
     */
    public boolean isOpen() {
        return accessMode!=NOT_OPEN;
    }

    /**
     * @return access type.
     */
    public int getAccessMode() {
        return accessMode;
    }

    /**
     * Check does the current database is opened.
     */
    public String getName() {
        return name;
    }

    /**
     * Check does the current database is opened.
     */
    public ObjectManager getObjectManager() {
        return objmanager;
    }

    /**
     * Open the named database with the specified access mode. Attempts to open
     * a database when it has already been opened will result in the throwing
     * of the exception DatabaseOpenException.
     * ODMG 3.0 implementation.
     *
     * @param name       the name of the database (or it's url).
     * @param accessMode the access mode, which should be one of the static
     *                   fields: OPEN_READ_ONLY, OPEN_READ_WRITE, or
     *                   OPEN_EXCLUSIVE.
     * @throws           DatabaseNotFoundException if the database doesn't
     *                   exist.
     */
    public void open(String name, int accessMode) throws ODMGException {
        if (isOpen())
            throw new DatabaseOpenException( "Database '" + this.name + "' has already been opened." );
        long cur = System.currentTimeMillis();
        this.accessMode = OPEN_EXCLUSIVE;
        this.name = name;
        // Check initializing of OODB
        OODB.init();
        OODB.getCurrent().bind( name, this );
        // Load all objects into the memory
        new ObjectSerializer().load( this );
        OODB.getCurrent().getLog().info( "[DB]   Database '" + name + "' opened. Loaded "
            + objmanager.getRootSize() + " persistent and "
            + objmanager.getNamedSize() + " named objects within "
            + (System.currentTimeMillis()-cur)/1000 + " sec." );
    }

    /**
     * Close the database. After you have closed a database, further attempts
     * to access objects in the database will cause the exception
     * DatabaseClosedException to be thrown.
     * ODMG 3.0 implementation.
     */
    public void close() throws ODMGException {
        if (isOpen()) {
            long cur = System.currentTimeMillis();
            new ObjectSerializer().save( this );
            OODB.getCurrent().unbind( name );
            accessMode = NOT_OPEN;
            OODB.getCurrent().getLog().info( "[DB]   Database '" + name + "' closed. Saved "
                + objmanager.getRootSize() + " persistent and "
                + objmanager.getNamedSize() + " named objects within "
                + (System.currentTimeMillis()-cur)/1000 + " sec." );
        }
    }

    /**
     * Adds transient object to the object manager.
     * Is called automatically on any MetaObject class creation within opened
     * transaction.
     *
     * @param object  object which should be added.
     * @throws        ObjectNameNotUniqueException if the persistent object
     *                with this OID already exist in this database.
     * @throws        OODBRuntimeException if the object is not of type
     *                MetaObject.
     */
    public void addTransientObject(Object object) throws ObjectNotUniqueException {
        // Check type of object
        OODB.isMetaObject(object);
        // Call object manager for this task
        objmanager.addTransientObject( (MetaObject)object );
    }

    /**
     *
     * Is called automatically within opened transaction on abort.
     *
     * @param object  the object to remove.
     * @throws        ObjectNameNotFoundException if the object of the
     *                specified name doesn't exist in this database.
     * @throws        OODBRuntimeException if the object is not of type
     *                MetaObject.
     */
    public void removeTransientObject(Object object) throws ObjectNotFoundException {
        // Check type of object
        OODB.isMetaObject(object);
        // Call object manager for this task
        objmanager.removeTransientObject( (MetaObject)object );
    }

    /**
     * Bind the named object for easy retrival.
     * Should be called manually by user within opened transaction.
     * ODMG 3.0 implementation.
     *
     * @param object  object which should be binded. Must be of type MetaObject
     *                otherwise OODBRuntimeException will be thrown.
     * @param name    string for name.
     * @throws        ObjectNameNotUniqueException if specified name already
     *                exists.
     * @throws        OODBRuntimeException if the object is not of type
     *                MetaObject.
     */
    public void bind(Object object, String name) throws ObjectNameNotUniqueException {
        // Check type of object
        OODB.isMetaObject(object);
        if (name == null)
            OODB.getCurrent().getLog().warn( "[DB]   [Database: "
                + this.name + "]\tYou are trying to bind named object which name=null. This will do nothing." );
        // Call object manager for this task
        objmanager.bind( (MetaObject)object, name );
    }

    /**
     * Get the named object by it's name.
     * Should be called manually by user within opened transaction.
     * ODMG 3.0 implementation.
     *
     * @param name    name of the binded object.
     * @return        an instance of the persistent object.
     * @throws        ObjectNameNotFoundException if the named object of the
     *                specified name doesn't exist in this database.
     */
    public Object lookup(String name) throws ObjectNameNotFoundException {
        if (!isOpen())
            throw new TransactionNotInProgressException( "Database '" + this.name + "' is not opened." );
        if (Transaction.getCurrent() == null)
            throw new TransactionNotInProgressException("Opened transaction not exists");
        // Call object manager for this task. Return copy with incremented version.
        MetaObject obj = objmanager.lookup(name);
        obj.setLock( Transaction.UPGRADE );
        // Register
        Transaction.getCurrent().registerTransientObject( obj );
        return obj;
    }

    /**
     * Disassociate a name with an object. It doesn't remove a persistent
     * object from database.
     * ODMG 3.0 implementation.
     *
     * @param name    name of the binded object.
     * @throws        ObjectNameNotFoundException if the named object of the
     *                specified name doesn't exist in this database.
     */
    public void unbind(String name) throws ObjectNameNotFoundException {
        // Call object manager for this task
        objmanager.unbind(name);
    }

    /**
     * Make a transient object durable in the database. It is executed in
     * the context of an open transaction. If the transaction in which this
     * method is executed commits, then the object is made durable. If the
     * transaction aborts, then the makePersistent operation is considered not
     * to have been executed, and the target object is again transient.
     * Is called automatically on commit within opened transaction.
     * ODMG 3.0 implementation.
     *
     * @param object  the object to make persistent.
     * @throws        ClassNotPersistenceCapableException is thrown if the
     *                implementation cannot make the object persistent because
     *                of the type of the object.
     * @throws        OODBRuntimeException if the object is not of type
     *                MetaObject.
     */
    public void makePersistent( Object object ) throws ObjectNotFoundException {
        // Check type of object
        OODB.isMetaObject(object);
        // Call object manager for this task
        objmanager.addPersistentObject( (MetaObject)object );
    }

    /**
     * Deletes an object from the database. It must be executed in the context
     * of an open transaction. If the transaction in which this method
     * is executed commits, then the object is removed from the database.
     * If the transaction aborts, then the deletePersistent operation is
     * considered not to have been executed, and the target object is again in
     * the database.
     * ODMG 3.0 implementation.
     *
     * @param object  the object to delete.
     * @throws        ObjectNotPersistent if the object is not persistent.
     *                of the type of the object.
     * @throws        OODBRuntimeException if the object is not of type
     *                MetaObject.
     */
    public void deletePersistent( Object object ) throws ObjectNotFoundException {
        // Check type of object
        OODB.isMetaObject(object);
        // Call object manager for this task
        objmanager.removePersistentObject( (MetaObject)object );
    }

    /**
     * Debug implementation.
     * Dump number of name binded objects, transient and persistent objects.
     */
    public void dump() {
        objmanager.dump();
    }

    /**
     * Is the transaction associated with this thread opened.
     */
    public boolean isTransactionOpened() {
        Transaction tr = Transaction.getCurrent();
        if (tr == null)
            return false;
        return true;
    }

}