package lv.tsi.oodb.odmg;

import org.odmg.*;

/**
 * Set.
 *
 * This class represents relationship 1:M
 * All values are unique and not ordered.
 * Equivalent to ODMG 3.0
 *   SET <Type>
 *
 * @author Alexey Vasilyev
 * @version 0.0.1 4 November 2000
 */
public class SetOfObject extends java.util.HashSet implements org.odmg.DSet {

    /**
     * Create instance of this class and reserve 5 entries in hashtable.
     */
    public SetOfObject() {
        super(5);
    }

    /**
     * UNION.
     * For example, we have two sets A(current) and B(otherSet).
     * A = (1,2,3,4). B = (2,3,4,5).
     *    A U B = (1,2,3,4,5)
     *
     * @return new DSet object which is the union of the two sets: the current and the given
     */
    public DSet	union( DSet otherSet ) {
        DSet set = (DSet)this.clone();
        set.addAll( otherSet );
        return set;
    }

    /**
     * INTERSECTION.
     * For example, we have two sets A(current) and B(otherSet).
     * A = (1,2,3,4). B = (2,3,4,5).
     *    A ^ B = (2,3,4)
     *
     * @return new DSet object which is the intersection of the two sets: the current and the given
     */
    public DSet	intersection( DSet otherSet ) {
        DSet set = (DSet)this.clone();
        set.retainAll( otherSet );
        return set;
    }

    /**
     * DIFFERENCE.
     * For example, we have two sets A(current) and B(otherSet).
     * A = (1,2,3,4). B = (2,3,4,5).
     *    A - B = (1)
     *
     * @return new DSet object which is the difference of the two sets: the current and the given
     */
    public DSet difference( DSet otherSet ) {
        DSet set = (DSet)this.clone();
        set.removeAll( otherSet );
        return set;
    }

    /**
     * SUBSET.
     * For example, we have two sets A(current) and B(otherSet).
     * A = (2,3). B = (1,2,3,4).
     *    A subsetOf B = true
     *
     * A = (3,5). B = (1,2,3,4).
     *    A subsetOf B = false
     *
     * @return true if this set is a subset of the set referenced by otherSet.
     */
    public boolean subsetOf( DSet otherSet ) {
        return otherSet.containsAll( this );
    }

    /**
     * SUPERSET.
     * For example, we have two sets A(current) and B(otherSet).
     * A = (1,2,3,4). B = (2,4).
     *    A supersetOf B = true
     *
     * A = (1,2,3,4). B = (1,2,3,4,5).
     *    A supersetOf B = false
     *
     * @return true if this set is a superset of the set referenced by otherSet.
     */
    public boolean supersetOf( DSet otherSet ) {
        return otherSet.subsetOf( this );
    }

    /**
     * PROPER SUBSET.
     * For example, we have two sets A(current) and B(otherSet).
     * A = (2,3). B = (1,2,3,4).
     *    A properSubsetOf B = true
     *
     * A = (1,2,3,4). B = (1,2,3,4).
     *    A properSubsetOf B = false
     *
     * @return true  if every element in this set is also in a specified otherSet and this
     *               set smaller than the otherSet.
     *
     */
    public boolean properSubsetOf( DSet otherSet ) {
        return otherSet.size() > this.size() && subsetOf( otherSet );
    }

    /**
     * PROPER SUPERSET.
     * For example, we have two sets A(current) and B(otherSet).
     * A = (1,2,3,4). B = (2,4).
     *    A supersetOf B = true
     *
     * A = (1,2,3,4). B = (1,2,3,4).
     *    A supersetOf B = false
     *
     * @return true  if every element in this set is also in a specified otherSet and this
     *               set smaller than the otherSet.
     *
     */
    public boolean properSupersetOf( DSet otherSet ) {
        return otherSet.properSubsetOf( this );
    }

// DCollection

    /**
     * @param predicate   OQL predicate. For example, "student=5".
     */
    public Object selectElement(String predicate) throws QueryInvalidException {
        // Not implemented yet
        throw new NotImplementedException ("OQL not supported");
    }

    public java.util.Iterator select(String predicate) throws QueryInvalidException {
        // Not implemented yet
        throw new NotImplementedException ("OQL not supported");
    }

    public DCollection query(String predicate) throws QueryInvalidException {
        // Not implemented yet
        throw new NotImplementedException ("OQL not supported");
    }

    public boolean existsElement(String predicate) throws QueryInvalidException {
        // Not implemented yet
        throw new NotImplementedException ("OQL not supported");
    }

}