package lv.tsi.oodb.odmg;

import lv.tsi.oodb.core.*;
import lv.tsi.oodb.util.Log;
import lv.tsi.oodb.OODB;
import org.odmg.*;

import java.util.*;

/**
 * Transaction implementation.
 *
 * This class provides the operations necessary to perform database
 * transactions. All access, creation, and modification of persistent objects
 * and their fields must be done within a transaction. Before performing any
 * database operations, a thread must explicitly create a transaction object or
 * associate itself with an existing transaction object (by calling join),
 * and that transaction must be open (through a call to begin). All subsequent
 * operations by the thread, including reads, writes, and lock acquisitions,
 * are done under the thread's current transaction. 
 *
 * A thread may only operate on its current transaction. For example,
 * a TransactionNotInProgressException is thrown if a thread attempts to begin,
 * commit, checkpoint, or abort a transaction prior to joining itself to that
 * transaction. 
 *
 * A transaction is either open or closed. A transaction is open if a call has
 * been made to begin, but no call has been made to commit or abort.
 * Once commit or abort is called, the transaction is closed. The method isOpen
 * can be called to determine the state of the transaction. 
 *
 * Read locks are implicitly obtained on objects as they are accessed. Write
 * locks are implicitly obtained as objects are modified. Transaction objects
 * are transient, they cannot be stored in the database. 
 *
 * @author  Alexey Vasilyev
 * @version 0.0.4 18 November 2000
 */
public class Transaction implements org.odmg.Transaction, Debug {

    /** Registry of not committed transient objects **/
    private Vector objects = new Vector();

    /** Currently associated database to this transaction **/
    private Database db;

    //public Transaction() {
    //    this( Database.getCurrent() );
    //}

    public Transaction(Database db) {
        this.db = db;
    }

    /**
     * Attach the caller's thread to this Transaction and detach the thread
     * from any former Transaction the thread may have been associated with.
     * ODMG 3.0 implementation.
     */
    public void join() {
    }

    /**
     * Detach the caller's thread from this Transaction, but do not attach the
     * thread to another Transaction.
     * ODMG 3.0 implementation.
     */
    public void leave() {
    }

    /**
     * Start a transaction.
     * ODMG 3.0 implementation.
     *
     * @throws           TransactionInProgressException is thrown on calling
     *                   begin multiple times on the same transaction object,
     *                   without an intervening call to commit or abort.
     * @throws           TransactionNotInProgressException can be thrown if
     *                   operations executed before a transaction has been
     *                   opened, or before reopening after a transaction is
     *                   aborted or committed, have undefined results.
     */
    public void begin() throws TransactionInProgressException {
        if (isOpen())
            throw new TransactionInProgressException( "Transaction already is in progress." );
        // Register current transaction
        OODB.getCurrent().getTransactionManager().register( this );
        Log.getCurrent().debug( "[TXN]  BEGIN" );
    }

    /**
     * Determine whether the transaction is open or not. A transaction is open
     * if a call has been made to begin, but a subsequent call to either commit
     * or abort has not been made.
     * ODMG 3.0 implementation.
     *
     * @return           true if the transaction is open, otherwise false.
     */
    public boolean isOpen() {
        return getCurrent() != null;
    }

    /**
     * Close the currently opened transaction.
     */
    private void close() {
        if (isOpen()) {
            objects = new Vector();
            OODB.getCurrent().getTransactionManager().unregister( this );
        }
    }

    /**
     * Commit and close the transaction. Calling commit commits to the database
     * all persistent object modifications within the transaction and releases
     * any locks held by the transaction. A persistent object modification is
     * an update of any field of an existing persistent object, or an update or
     * creation of a new named object in the database. If a persistent object
     * modification results in a reference from an existing persistent object
     * to a transient object, the transient object is moved to the database,
     * and all references to it updated accordingly. Note that the act of
     * moving a transient object to the database may create still more
     * persistent references to transient objects, so its referents must be
     * examined and moved as well. This process continues until the database
     * contains no references to transient objects, a condition that is
     * guaranteed as part of transaction commit. Committing a transaction does
     * not remove from memory transient objects created during the transaction.
     *
     * This will copy transaction object table to the root object table.
     */
    public void commit() throws ODMGRuntimeException {
        if (isOpen()) {
            commit0();
            close();
            //  -----------------------------------------------
            // | Phew... all is ok.                            |
            //  -----------------------------------------------
            Log.getCurrent().debug( "[TXN]  COMMITED" );
        }
    }

    /**
     * Abort and close the transaction. Calling abort abandons all persistent
     * object modifications and releases the associated locks. Aborting
     * a transaction does not restore the state of modified transient objects.
     * ODMG 3.0 implementation.
     */
    public void abort() {
        if (isOpen()) {
            //  -----------------------------------------------
            //  -----------------------------------------------
            try {
                Iterator i = objects.iterator();
                while (i.hasNext()) {
                    // This will also tries to remove name binded objects
                    db.removeTransientObject( (MetaObject)i.next() );
                }
            } catch (ODMGRuntimeException e) {
            } finally {
                close();
            }
            Log.getCurrent().debug( "[TXN]  ABORTED" );
        }
    }

    /**
     * Commit the transaction, but reopen the transaction, retaining all locks.
     * Calling checkpoint commits persistent object modifications made within
     * the transaction since the last checkpoint to the database.
     * The transaction retains all locks it held on those objects at the time
     * the checkpoint was invoked.
     * ODMG 3.0 implementation.
     */
    public void checkpoint() {
        if (isOpen()) {
            commit0();
            Log.getCurrent().debug( "[TXN]  CHECKPOINTED" );
        }
    }

    /**
     * Commit the transaction, but don't close it.
     */
    private final void commit0() {
        int cur = 0;
        Iterator i = objects.iterator();
        //  -----------------------------------------------
        // | First commit phase.                           |
        //  -----------------------------------------------
        try {
            while (i.hasNext()) {
                db.makePersistent( (MetaObject)i.next() );
                cur++;
            }
        } catch (ODMGRuntimeException e) {
            // makePersistent produced an error. Rollback all changes.
            try {
                i = objects.iterator();
                while (cur--!=0)
                    // This must be redesigned !
                    db.deletePersistent( (MetaObject)i.next() );
            } catch (ODMGRuntimeException ex) {
            } finally {
                close();
                throw e;
            }
        }
        //  -----------------------------------------------
        // | Second commit phase.                          |
        //  -----------------------------------------------
        try {
            i = objects.iterator();
            while (i.hasNext()) {
                MetaObject obj = (MetaObject)i.next();
                db.removeTransientObject( obj );
                // Set READ and NORMAL mode
                obj.setOk();
            }
        } catch (ODMGRuntimeException e) {
            // Something wrong
        }
        // All is ok.
    }

    /**
     * Upgrade the lock on the given object to the given lock mode.
     * The call has no effect if the object's current lock is already at or
     * above that level of lock mode.
     * ODMG 3.0 implementation.
     *
     * @param object     the object to acquire a lock on..
     * @param lockMode   the lock mode to acquire. The lock modes are READ,
     *                   UPGRADE, and WRITE.
     * @throws           LockNotGrantedException ss thrown if the given lock
     *                   mode could not be acquired.
     */
    public void lock(Object object, int lockMode) throws LockNotGrantedException {
        // Check type of object
        OODB.isMetaObject(object);
        // Setting lock
        //MetaObject meta = (MetaObject)object;
        //if (meta.getLockMode()
        
    }

    /**
     * Upgrade the lock on the given object to the given lock mode.
     * Method tryLock is the same as lock except it returns a boolean
     * indicating whether the lock was granted instead of generating
     * an exception.
     * ODMG 3.0 implementation.
     *
     * @param object     the object to acquire a lock on..
     * @param lockMode   the lock mode to acquire. The lock modes are READ,
     *                   UPGRADE, and WRITE.
     * @throws           LockNotGrantedException ss thrown if the given lock
     *                   mode could not be acquired.
     */
    public boolean tryLock( Object object, int lockMode ) {
        // Fake
        return true;
    }

    /**
     * Put the given transient object to the transaction object table and
     * to the database's object manager.
     * Each MetaObject calls this method on its creating.
     * Not an ODMG feature.
     */
    public void registerTransientObject( MetaObject object ) throws ObjectNotFoundException {
        objects.add( object );
        // Call database
        db.addTransientObject( object );
    }

    /**
     * @return   Database object associated whith this transaction.
     */
    public Database getDatabase() {
        return db;
    }

    /**
     * Print debug information such as number of transient objects.
     */
    public void dump() {
        Log.getCurrent().debug( "[TXN]  Objects#: " + objects.size() );
    }

///////////////////////////
// Static methods
//
    /**
     * Get the current opened transaction of the current thread.
     */
    public static Transaction getCurrent() {
        return getCurrent( Thread.currentThread() );
    }

    /**
     * Get the current opened transaction of the current thread.
     *
     * @param th  thread which
     */
    public static Transaction getCurrent( Thread th ) {
        return TransactionManager.getTransaction( th );
    }

    /**
     * String representation of the given lock mode.
     */
    public static String lockToString( int lockMode ) {
        switch (lockMode) {
            case READ:    return "READ";
            case UPGRADE: return "UPGRADE";
            case WRITE:   return "WRITE";
        }
        return "UNKNOWN";
    }

}