package lv.tsi.oodb.util;

import lv.tsi.oodb.OODB;

import java.util.Date;
import java.text.*;
import java.io.*;

/**
 * Loging system.
 * Can log into the stdout and into the file.
 * It has 4 log levels: debug, info, warning, error.
 * Functionality of this class is similar to Apache (http://www.apache.org)
 * web server's log system.
 *
 * @version 0.0.5 4 December 2000
 * @author Alexey Vasilyev
 */
public class Log {

    public final static int DBG = 0;
    public final static int INF = 1;
    public final static int WRN = 2;
    public final static int ERR = 3;

    private int              level = INF;
    private PrintWriter      out   = new PrintWriter( System.out );

    private DateFormat form = DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.MEDIUM);
    //private SimpleDateFormat form  = new SimpleDateFormat(DateFormat.FULL
 //DateFormat.SHORT, Locale.getDefault()

    /**
     * @param filename  the file to loging into. If null or incorrect then
     *                  stdout is used.
     * @param logLevel  can be Log.DBG | Log.INF | Log.WRN | Log.ERR
     */
    public Log (String filename, int logLevel) {
        init( filename, logLevel );
    }

    /**
     * Open connection to the file if the filename specified (not null) or
     * open stdout connection.
     *
     * @param filename  the file to loging into. If null or incorrect then
     *                  stdout is used.
     * @param logLevel  can be Log.DBG | Log.INF | Log.WRN | Log.ERR
     */
    public void init (String filename, int logLevel) {
        switch (logLevel) {
            case INF:
            case WRN:
            case ERR:
            case DBG: level = logLevel;
                      break;
            default:  level = INF;
        }
        if (filename != null)
           try {
               out = new PrintWriter( new FileWriter( filename, true ), true );
           } catch (IOException e) {
               System.out.println( "Log::init() " + e );
               out = new PrintWriter( System.out, true );
           }
        else
           out = new PrintWriter( System.out, true );
    }

    /**
     * Close connection to the file.
     */
    public void close() {
        out.close();
    }

    /**
     * Log the given message to the stream. To log debug messages log level
     * must be Log.DBG.
     *
     * @param msg  text to log.
     */
    public synchronized void debug (String msg) {
        if ( DBG == level )
            out.println( "[" + form.format(new Date()) + "] [DBG] " + msg );
    }

    /**
     * Log the given information message to the stream. To log information messages
     * log level must be Log.DBG or Log.INF.
     *
     * @param msg  text to log.
     */
    public synchronized void info (String msg) {
        if ( INF >= level )
            out.println( "[" + form.format(new Date()) + "] [INF] " + msg );
    }

    /**
     * Log the given warning message to the stream. To log warning messages log level
     * must be Log.DBG, Log.INF or Log.WRN.
     *
     * @param msg  text to log.
     */
    public synchronized void warn (String msg) {
        if ( WRN >= level )
            out.println( "[" + form.format(new Date()) + "] [WRN] " + msg );
    }

    /**
     * Log the given error message to the stream. Error messages allways are loging.
     *
     * @param msg  text to log.
     */
    public synchronized void error (String msg) {
        if ( ERR >= level )
            out.println( "[" + form.format(new Date()) + "] [ERR] " + msg );
    }

///////////////////////////
// Static methods
//
    /**
     * @param level  string presentation of the log level description.
     *               Can be "debug" | "info" | "warning" | "error"
     *
     * @return Log.DBG | Log.INF | Log.WRN | Log.ERR
     */
    public static int getLogLevel( String level ) {
        if (level == null)
            return INF;
        else if (level.equals("debug"))
            return DBG;
        else if (level.equals("info"))
            return INF;
        else if (level.equals("warning"))
            return WRN;
        else if (level.equals("error"))
            return ERR;
        return INF;
    }

    /**
     * @return the currently opened instance of Log class.
     */
    public static Log getCurrent() {
        return OODB.getCurrent().getLog();
    }

}